from typing import Tuple
import numpy


TransformationMatrix = numpy.ndarray  # 4x4 matrix
CameraLocationType = Tuple[float, float, float]
CameraRotationType = Tuple[float, float]

RGBColour = Tuple[float, float, float]
